
####################################################################################################
#
# PySpice - A Spice package for Python
# Copyright (C) 2019 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from pathlib import Path
import os
import shutil

from invoke import task
 # import sys

from .clean import flycheck as _clean_flycheck
from .release import update_git_sha as _update_git_sha

####################################################################################################

PYSPICE_SOURCE_PATH = Path(__file__).resolve().parents[1]

EXAMPLES_PATH = PYSPICE_SOURCE_PATH.joinpath('examples')

SPHINX_PATH = PYSPICE_SOURCE_PATH.joinpath('doc', 'sphinx')
BUILD_PATH = SPHINX_PATH.joinpath('build')
RST_SOURCE_PATH = SPHINX_PATH.joinpath('source')
RST_API_PATH = RST_SOURCE_PATH.joinpath('api')
RST_EXAMPLES_PATH = RST_SOURCE_PATH.joinpath('examples')

####################################################################################################

@task
def clean_build(ctx):
    # ctx.run('rm -rf {}'.format(BUILD_PATH))
    if BUILD_PATH.exists():
        shutil.rmtree(BUILD_PATH)

####################################################################################################

@task
def clean_api(ctx):
    # ctx.run('rm -rf {}'.format(RST_API_PATH))
    if RST_API_PATH.exists():
        shutil.rmtree(RST_API_PATH)

@task(_update_git_sha, _clean_flycheck, clean_api)
def make_api(ctx):
    print('\nGenerate RST API files')
    ctx.run('pyterate-rst-api {0.Package}'.format(ctx))
    run_sphinx(ctx)

####################################################################################################

@task(_update_git_sha)
def make_examples(ctx, clean=False, no_html=False, force=False):

    # Regenerate from scratch
    if clean and RST_EXAMPLES_PATH.exists():
        shutil.rmtree(RST_EXAMPLES_PATH)

    # pyterate --skip-external-figure --skip-figure
    # PYTHONPATH=$PWD/examples/:${PYTHONPATH}
    # PySpiceLogLevel=WARNING

    os.environ['PySpiceLibraryPath'] = str(EXAMPLES_PATH.joinpath('libraries'))
    os.environ['PySpiceLogLevel'] = 'ERROR'  # set logging level

    setting_path = EXAMPLES_PATH.joinpath('Settings.py')
    # subprocess.run(('pyterate',
    #                 '--config', str(setting_path))
    # )
    print('Generate RST examples files')
    command = [
        'pyterate',
        '--config={}'.format(setting_path),
    ]
    if force:
        command.append('--force')
    ctx.run(' '.join(command))

    if not no_html:
        run_sphinx(ctx)

####################################################################################################

@task
def run_sphinx(ctx):
    print('\nRun Sphinx')
    working_path = SPHINX_PATH
    # subprocess.run(('make-html'), cwd=working_path)
    # --clean
    with ctx.cd(str(working_path)):
        ctx.run('make-html')

####################################################################################################

@task
def make_readme(ctx):
    # File "/usr/bin/rst2html", line 17, in <module>
    # from docutils.core import publish_cmdline, default_description
    # ModuleNotFoundError: No module named 'docutils'
    from setup_data import long_description
    with open('README.rst', 'w') as fh:
        fh.write(long_description)
    # import subprocess
    # subprocess.call(('rst2html', 'README.rst', 'README.html'))
    ctx.run('rst2html5.py README.rst README.html')

####################################################################################################

@task
def update_authors(ctx):
    # Keep authors in the order of appearance and use awk to filter out dupes
    ctx.run("git log --format='- %aN <%aE>' --reverse | awk '!x[$0]++' > AUTHORS")

####################################################################################################

@task
def publish(ctx):
    from .SECRET_CONFIG import SSH_CONFIG
    import PySpice
    release = PySpice.__version__
    version = '.'.join(release.split('.')[:2])
    command_template = (
        'rsync'
        ' -av -c --delete'
        ' --exclude="*~" --delete-excluded'
        ' -e "ssh -p {ssh_port}" {src_path}/ {ssh_user}@{ssh_host}:{ssh_path}/releases/v{version}/'
    )
    command = command_template.format(
        src_path=BUILD_PATH.joinpath('html'),
        version=version,
        **SSH_CONFIG)
    print(command)
    ctx.run(command)

####################################################################################################

@task
def xdg_open(ctx):
    ctx.run('xdg-open doc/sphinx/build/html/index.html')



#r# ================
#r#  CEM Simulation
#r# ================

#r# This example show a CEM simulation.

# Fixme: retrieve PDF reference and complete

####################################################################################################

import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Probe.Plot import plot
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

from HP54501A import HP54501A

#f# literal_include('HP54501A.py')

####################################################################################################

circuit = Circuit('HP54501A CEM')
circuit.include(spice_library['1N4148'])
diode_model = '1N4148'
ac_line = circuit.AcLine('input', 'input', circuit.gnd, rms_voltage=230@u_V, frequency=50@u_Hz)
# circuit.subcircuit(HP54501A(diode_model='1N4148'))
# circuit.X('hp54501a', 'HP54501A', 'input', circuit.gnd)
circuit.C(1, 'input', circuit.gnd, 1@u_uF)
circuit.X('D1', diode_model, 'line_plus', 'top')
circuit.X('D2', diode_model, 'scope_ground', 'input')
circuit.X('D3', diode_model, circuit.gnd, 'top')
circuit.X('D4', diode_model, 'scope_ground', circuit.gnd)
circuit.R(1, 'top', 'output', 10@u_Ω)
circuit.C(2, 'output', 'scope_ground', 50@u_uF)
circuit.R(2, 'output', 'scope_ground', 900@u_Ω)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.transient(step_time=ac_line.period/100, end_time=ac_line.period*3)

figure, ax = plt.subplots(figsize=(20, 6))

ax.plot(analysis.input)
ax.plot(analysis.Vinput)
ax.plot(analysis.output - analysis.scope_ground)
ax.legend(('Vin [V]', 'I [A]'), loc=(.8,.8))
ax.grid()
ax.set_xlabel('t [s]')
ax.set_ylabel('[V]')

plt.show()

#f# save_figure('figure', 'hp54501a-cem.png')



#r# ====================
#r#  Bipolar Transistor
#r# ====================

#r# This example shows how to simulate the characteristic curves of a bipolar transistor.

# Fixme: Complete

####################################################################################################

import numpy as np
import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Probe.Plot import plot
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

figure, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(20, 10))

####################################################################################################

#r# We define a basic circuit to drive an NPN transistor (2n2222a) using two voltage sources.

#f# circuit_macros('transistor.m4')

circuit = Circuit('Transistor')

Vbase = circuit.V('base', '1', circuit.gnd, 1@u_V)
circuit.R('base', 1, 'base', 1@u_kΩ)
Vcollector = circuit.V('collector', '2', circuit.gnd, 0@u_V)
circuit.R('collector', 2, 'collector', 1@u_kΩ)
# circuit.BJT(1, 'collector', 'base', circuit.gnd, model='generic')
# circuit.model('generic', 'npn')
circuit.include(spice_library['2n2222a'])
circuit.BJT(1, 'collector', 'base', circuit.gnd, model='2n2222a')

#r# We plot the base-emitter diode curve :math:`Ib = f(Vbe)` using a DC sweep simulation.

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.dc(Vbase=slice(0, 3, .01))

ax1.plot(analysis.base, u_mA(-analysis.Vbase)) # Fixme: I_Vbase
ax1.axvline(x=.65, color='red')
ax1.legend(('Base-Emitter Diode curve',), loc=(.1,.8))
ax1.grid()
ax1.set_xlabel('Vbe [V]')
ax1.set_ylabel('Ib [mA]')

####################################################################################################

#r# We will now replace the base's voltage source by a current source in the previous circuit.

circuit = Circuit('Transistor')
Ibase = circuit.I('base', circuit.gnd, 'base', 10@u_uA) # take care to the orientation
Vcollector = circuit.V('collector', 'collector', circuit.gnd, 5)
# circuit.BJT(1, 'collector', 'base', circuit.gnd, model='generic')
# circuit.model('generic', 'npn')
circuit.include(spice_library['2n2222a'])
circuit.BJT(1, 'collector', 'base', circuit.gnd, model='2n2222a')

# Fixme: ngspice doesn't support multi-sweep ???
#   it works in interactive mode

#?# simulator = circuit.simulator(temperature=25, nominal_temperature=25)
#?# analysis = simulator.dc(Vcollector=slice(0, 5, .1), Ibase=slice(micro(10), micro(100), micro(10)))
#?# 0 v(i-sweep)    voltage # Vcollector in fact
#?# 1 v(collector)  voltage
#?# 2 v(base)       voltage
#?# 3 i(vcollector) current
#?# 0.00000000e+00,   1.00000000e-01,   2.00000000e-01, 3.00000000e-01,   4.00000000e-01,   5.00000000e-01, 6.00000000e-01,   7.00000000e-01,   8.00000000e-01, 9.00000000e-01
#?# 0.00000000e+00,   1.00000000e-01,   2.00000000e-01, 3.00000000e-01,   4.00000000e-01,   5.00000000e-01, 6.00000000e-01,   7.00000000e-01,   8.00000000e-01, 9.00000000e-01
#?# 6.50478604e-01,   7.40522920e-01,   7.68606463e-01, 7.69192913e-01,   7.69049191e-01,   7.69050844e-01, 7.69049584e-01,   7.69049559e-01,   7.69049559e-01, 7.69049559e-01
#?# 9.90098946e-06,  -3.15540984e-04,  -9.59252614e-04, -9.99134834e-04,  -9.99982226e-04,  -1.00005097e-03, -1.00000095e-03,  -9.99999938e-04,  -9.99999927e-04, -9.99999937e-04
#?#
#?# analysis = simulator.dc(Vcollector=slice(0, 10, .1))
#?# 0 v(v-sweep)      voltage
#?# 1 v(collector)    voltage
#?# 2 v(base)         voltage
#?# 3 i(vcollector)   current
#?#
#?# analysis = simulator.dc(Ibase=slice(micro(10), micro(100), micro(10)))
#?# 0 v(i-sweep)      voltage
#?# 1 v(collector)    voltage
#?# 2 v(base)         voltage
#?# 3 i(vcollector)   current

ax2.grid()
# ax2.legend(('Ic(Vce, Ib)',), loc=(.5,.5))
ax2.set_xlabel('Vce [V]')
ax2.set_ylabel('Ic [mA]')
ax2.axvline(x=.2, color='red')

ax3.grid()
# ax3.legend(('beta(Vce)',), loc=(.5,.5))
ax3.set_xlabel('Vce [V]')
ax3.set_ylabel('beta')
ax3.axvline(x=.2, color='red')

for base_current in np.arange(0, 100, 10):
    base_current = base_current@u_uA
    Ibase.dc_value = base_current
    simulator = circuit.simulator(temperature=25, nominal_temperature=25)
    analysis = simulator.dc(Vcollector=slice(0, 5, .01))
    # add ib as text, linear and saturate region
    # Plot Ic = f(Vce)
    ax2.plot(analysis.collector, u_mA(-analysis.Vcollector))
    # Plot β = Ic / Ib = f(Vce)
    ax3.plot(analysis.collector, -analysis.Vcollector/float(base_current))
    # trans-resistance U = RI   R = U / I = Vce / Ie
    # ax3.plot(analysis.collector, analysis.sweep/(float(base_current)-analysis.Vcollector))
    # Fixme: sweep is not so explicit

#r# Let plot :math:`Ic = f(Ib)`

ax4.grid()
ax4.set_xlabel('Ib [uA]')
ax4.set_ylabel('Ic [mA]')

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.dc(Ibase=slice(0, 100e-6, 10e-6))
# Fixme: sweep
ax4.plot(analysis.sweep*1e6, u_mA(-analysis.Vcollector), 'o-')
ax4.legend(('Ic(Ib)',), loc=(.1,.8))

####################################################################################################

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'transistor-plot.png')



# -*- coding: utf-8 -*-

####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module implements high level elements built on top of Spice elements."""

# Fixme: check NgSpice for discrepancies

####################################################################################################

from ..Math import rms_to_amplitude, amplitude_to_rms
from ..Tools.StringTools import join_list, join_dict, str_spice, str_spice_list
from ..Unit import as_s, as_V, as_A, as_Hz
from .BasicElement import VoltageSource, CurrentSource

####################################################################################################

class SourceMixinAbc:
    AS_UNIT = None

####################################################################################################

class VoltageSourceMixinAbc:
    AS_UNIT = as_V

####################################################################################################

class CurrentSourceMixinAbc:
    AS_UNIT = as_A

####################################################################################################

class SinusoidalMixin(SourceMixinAbc):

    r"""This class implements a sinusoidal waveform.

    +------+----------------+---------------+-------+
    | Name + Parameter      + Default Value + Units |
    +------+----------------+---------------+-------+
    | Vo   + offset         +               + V, A  |
    +------+----------------+---------------+-------+
    | Va   + amplitude      +               + V, A  |
    +------+----------------+---------------+-------+
    | f    + frequency      + 1 / TStop     + Hz    |
    +------+----------------+---------------+-------+
    | Td   + delay          + 0.0           + sec   |
    +------+----------------+---------------+-------+
    | Df   + damping factor + 0.01          + 1/sec |
    +------+----------------+---------------+-------+

    The shape of the waveform is described by the following formula:

    .. math::

        V(t) = \begin{cases}
          V_o & \text{if}\ 0 \leq t < T_d, \\
          V_o + V_a e^{-D_f(t-T_d)} \sin\left(2\pi f (t-T_d)\right) & \text{if}\ T_d \leq t < T_{stop}.
        \end{cases}

    Spice syntax::

        SIN ( Voffset Vamplitude Freq Tdelay DampingFactor )

    Public Attributes:

      :attr:`ac_magnitude`

      :attr:`amplitude`

      :attr:`damping_factor`

      :attr:`dc_offset`

      :attr:`delay`

      :attr:`frequency`

      :attr:`offset`

    """

    ##############################################

    def __init__(self,
                 dc_offset=0,
                 ac_magnitude=1,
                 offset=0, amplitude=1, frequency=50,
                 delay=0, damping_factor=0):

        self.dc_offset = self.AS_UNIT(dc_offset)
        self.ac_magnitude = self.AS_UNIT(ac_magnitude)
        self.offset = self.AS_UNIT(offset)
        self.amplitude = self.AS_UNIT(amplitude)
        self.frequency = as_Hz(frequency)   # Fixme: protect by setter?
        self.delay = as_s(delay)
        self.damping_factor = as_Hz(damping_factor)

    ##############################################

    @property
    def rms_voltage(self):
        # Fixme: ok ???
        return amplitude_to_rms(self.amplitude * self.ac_magnitude)

    ##############################################

    @property
    def period(self):
        return self.frequency.period

    ##############################################

    def format_spice_parameters(self):
        sin_part = join_list((self.offset, self.amplitude, self.frequency, self.delay, self.damping_factor))
        return join_list((
            'DC {} AC {}'.format(*str_spice_list(self.dc_offset, self.ac_magnitude)),
            'SIN({})'.format(sin_part),
        ))

####################################################################################################

class PulseMixin(SourceMixinAbc):

    """This class implements a pulse waveform.

    Nomenclature:

    +--------+---------------+---------------+-------+
    | Name   + Parameter     + Default Value + Units |
    +--------+---------------+---------------+-------+
    | V1     + initial value +               + V, A  |
    +--------+---------------+---------------+-------+
    | V2     + pulsed value  +               + V, A  |
    +--------+---------------+---------------+-------+
    | Td     + delay time    + 0.0           + sec   |
    +--------+---------------+---------------+-------+
    | Tr     + rise time     + Tstep         + sec   |
    +--------+---------------+---------------+-------+
    | Tf     + fall time     + Tstep         + sec   |
    +--------+---------------+---------------+-------+
    | Pw     + pulse width   + Tstop         + sec   |
    +--------+---------------+---------------+-------+
    | Period + period        + Tstop         + sec   |
    +--------+---------------+---------------+-------+
    | Phase  + phase         + 0.0           + sec   |
    +--------+---------------+---------------+-------+

    Phase is only possible when XSPICE is enabled

    Spice Syntax::

        PULSE ( V1 V2 Td Tr Tf Pw Period Phase )

    A single pulse so specified is described by the following table:

    +-------------+-------+
    | Time        | Value |
    +-------------+-------+
    | 0           | V1    |
    +-------------+-------+
    | Td          | V1    |
    +-------------+-------+
    | Td+Tr       | V2    |
    +-------------+-------+
    | Td+Tr+Pw    | V2    |
    +-------------+-------+
    | Td+Tr+Pw+Tf | V1    |
    +-------------+-------+
    | Tstop       | V1    |
    +-------------+-------+

    Note: default value in Spice for rise and fall time is the simulation transient step, pulse
    width and period is the simulation stop time.

    Public Attributes:

      :attr:`delay_time`

      :attr:`fall_time`

      :attr:`initial_value`

      :attr:`period`

      :attr:`phase`

      :attr:`pulse_width`

      :attr:`pulsed_value`

      :attr:`rise_time`

    """

    ##############################################

    def __init__(self,
                 initial_value, pulsed_value,
                 pulse_width, period,
                 delay_time=0, rise_time=0, fall_time=0,
                 phase=None,
                 dc_offset=0):

        # Fixme: default
        #  rise_time, fall_time = Tstep
        #  pulse_width, period = Tstop

        self.dc_offset = self.AS_UNIT(dc_offset)   # Fixme: -> SourceMixinAbc
        self.initial_value = self.AS_UNIT(initial_value)
        self.pulsed_value = self.AS_UNIT(pulsed_value)
        self.delay_time = as_s(delay_time)
        self.rise_time = as_s(rise_time)
        self.fall_time = as_s(fall_time)
        self.pulse_width = as_s(pulse_width)
        self.period = as_s(period)   # Fixme: protect by setter?

        # XSPICE
        if phase is not None:
            self.phase = as_s(phase)
        else:
            self.phase = None

        # # Fixme: to func?
        # # Check parameters
        # found_none = False
        # for parameter in ('rise_time', 'fall_time', 'pulse_width', 'period'):
        #     parameter_value = getattr(self, parameter)
        #     if found_none:
        #         if parameter_value is not None:
        #             raise ValueError("Parameter {} is set but some previous parameters was not set".format(parameter))
        #     else:
        #         found_none = parameter_value is None

    ##############################################

    @property
    def frequency(self):
        return self.period.frequency

    ##############################################

    def format_spice_parameters(self):

        # if DC is not provided, ngspice complains
        #   Warning: vpulse: no DC value, transient time 0 value used

        # Fixme: to func?
        return join_list((
            'DC {}'.format(str_spice(self.dc_offset)),
            'PULSE(' +
            join_list((self.initial_value, self.pulsed_value, self.delay_time,
                       self.rise_time, self.fall_time, self.pulse_width, self.period,
                       self.phase)) +
            ')'))

####################################################################################################

class ExponentialMixin(SourceMixinAbc):

    r"""This class implements a Exponential waveform.

    Nomenclature:

    +------+--------------------+---------------+-------+
    | Name + Parameter          + Default Value + Units |
    +------+--------------------+---------------+-------+
    | V1   + Initial value      +               + V, A  |
    +------+--------------------+---------------+-------+
    | V2   + pulsed value       +               + V, A  |
    +------+--------------------+---------------+-------+
    | Td1  + rise delay time    + 0.0           + sec   |
    +------+--------------------+---------------+-------+
    | tau1 + rise time constant + Tstep         + sec   |
    +------+--------------------+---------------+-------+
    | Td2  + fall delay time    + Td1+Tstep     + sec   |
    +------+--------------------+---------------+-------+
    | tau2 + fall time constant + Tstep         + sec   |
    +------+--------------------+---------------+-------+

    Spice Syntax::

        EXP ( V1 V2 TD1 TAU1 TD2 TAU2 )

    The shape of the waveform is described by the following formula:

    Let V21 = V2 - V1 and V12 = V1 - V2.

    .. math::

        V(t) = \begin{cases}
          V_1 & \text{if}\ 0 \leq t < T_{d1}, \\
          V_1 + V_{21} ( 1 − e^{-\frac{t-T_{d1}}{\tau_1}} )
          & \text{if}\ T_{d1} \leq t < T_{d2}, \\
          V_1 + V_{21} ( 1 − e^{-\frac{t-T_{d1}}{\tau_1}} ) + V_{12} ( 1 − e^{-\frac{t-T_{d2}}{\tau_2}} )
          & \text{if}\ T_{d2} \leq t < T_{stop}
        \end{cases}

    """

    ##############################################

    def __init__(self,
                 initial_value, pulsed_value,
                 rise_delay_time=.0, rise_time_constant=None,
                 fall_delay_time=None, fall_time_constant=None):

        # Fixme: default

        self.initial_value = self.AS_UNIT(initial_value)
        self.pulsed_value = self.AS_UNIT(pulsed_value)
        self.rise_delay_time = as_s(rise_delay_time)
        self.rise_time_constant = as_s(rise_time_constant)
        self.fall_delay_time = as_s(fall_delay_time)
        self.fall_time_constant = as_s(fall_time_constant)

    ##############################################

    def format_spice_parameters(self):
        # Fixme: to func?
        return ('EXP(' +
                join_list((self.initial_value, self.pulsed_value,
                           self.rise_delay_time, self.rise_time_constant,
                           self.fall_delay_time, self.fall_time_constant,
                           )) +
                ')')

####################################################################################################

class PieceWiseLinearMixin(SourceMixinAbc):

    r"""This class implements a Piece-Wise Linear waveform.

    Spice Syntax::

        PWL( T1 V1 <T2 V2 T3 V3 T4 V4 ... > ) <r=value> <td=value>

    Each pair of values (Ti , Vi) specifies that the value of the source is Vi (in Volts or Amps) at
    time = Ti . The value of the source at intermediate values of time is determined by using linear
    interpolation on the input values. The parameter r determines a repeat time point. If r is not
    given, the whole sequence of values (Ti , Vi ) is issued once, then the output stays at its
    final value. If r = 0, the whole sequence from time = 0 to time = Tn is repeated forever. If r =
    10ns, the sequence between 10ns and 50ns is repeated forever. the r value has to be one of the
    time points T1 to Tn of the PWL sequence. If td is given, the whole PWL sequence is delayed by a
    delay time time = td. The current source still needs to be patched, td and r are not yet
    available.

    `values` should be given as a list of (`Time`, `Value`)-tuples, e.g.::

        PieceWiseLinearVoltageSource(
            circuit,
            'pwl1', '1', '0',
            values=[(0, 0), (10@u_ms, 0), (11@u_ms, 5@u_V), (20@u_ms, 5@u_V)],
        )

    """

    ##############################################

    def __init__(self, values, repeat_time=None, delay_time=None, dc=None):
        self.values = sum(([as_s(t), self.AS_UNIT(x)] for (t, x) in values), [])
        self.repeat_time = as_s(repeat_time, none=True)
        self.delay_time = as_s(delay_time, none=True)
        self.dc = self.AS_UNIT(dc, none=True)

    ##############################################

    def format_spice_parameters(self):

        # Fixme: to func?

        d = {}
        if self.repeat_time is not None:
            d["r"] = self.repeat_time
        if self.delay_time is not None:
            d["td"] = self.delay_time

        _ = ""
        if self.dc is not None:
            _ += "DC {} ".format(str_spice(self.dc))
        _ += "PWL(" + join_list(self.values)
        if d:
            _ += " " + join_dict(d)   # OrderedDict(
        _ += ")"

        return _

####################################################################################################

class SingleFrequencyFMMixin(SourceMixinAbc):

    r"""This class implements a Single-Frequency FM waveform.

    Spice Syntax::

        SFFM (VO VA FC MDI FS )

   +------+-------------------+---------------+-------+
   | Name + Parameter         + Default Value + Units |
   +------+-------------------+---------------+-------+
   | Vo   + offset            +               + V, A  |
   +------+-------------------+---------------+-------+
   | Va   + amplitude         +               + V, A  |
   +------+-------------------+---------------+-------+
   | Fc   + carrier frequency + 1 / Tstop     + Hz    |
   +------+-------------------+---------------+-------+
   | Mdi  + modulation index  +               +       |
   +------+-------------------+---------------+-------+
   | Fs   + signal frequency  + 1 / Tstop     + Hz    |
   +------+-------------------+---------------+-------+

    The shape of the waveform is described by the following equation:

    .. math::

        V(t) = V_o + V_a \sin (2\pi F_c\, t + M_{di} \sin (2\pi F_s\,t))

    """

    ##############################################

    def __init__(self, offset, amplitude, carrier_frequency, modulation_index, signal_frequency):
        self.offset = self.AS_UNIT(offset)
        self.amplitude = self.AS_UNIT(amplitude)
        self.carrier_frequency = as_Hz(carrier_frequency)
        self.modulation_index = modulation_index
        self.signal_frequency = as_Hz(signal_frequency)

    ##############################################

    def format_spice_parameters(self):
        # Fixme: to func?
        return ('SFFM(' +
                join_list((self.offset, self.amplitude, self.carrier_frequency,
                           self.modulation_index, self.signal_frequency)) +
                ')')

####################################################################################################

class AmplitudeModulatedMixin(SourceMixinAbc):

    r"""This class implements a Amplitude Modulated source.

    +------+----------------------+---------------+-------+
    | Name + Parameter            + Default Value + Units |
    +------+----------------------+---------------+-------+
    | Vo   + offset               +               + V, A  |
    +------+----------------------+---------------+-------+
    | Va   + amplitude            +               + V, A  |
    +------+----------------------+---------------+-------+
    | Mf   + modulating frequency +               + Hz    |
    +------+----------------------+---------------+-------+
    | Fc   + carrier frequency    + 1 / Tstop     + Hz    |
    +------+----------------------+---------------+-------+
    | Td   + signal delay         +               + s     |
    +------+----------------------+---------------+-------+

    Spice Syntax::

        AM(VA VO MF FC TD)

    The shape of the waveform is described by the following equation:

    .. math::

        V(t) = V_a (V_o + \sin (2\pi M_f\,t)) \sin (2\pi F_c\,t)

    """

    ##############################################

    def __init__(self, offset, amplitude, modulating_frequency, carrier_frequency, signal_delay):

        # Fixme: default

        self.offset = self.AS_UNIT(offset)
        self.amplitude = self.AS_UNIT(amplitude)
        self.carrier_frequency = as_Hz(carrier_frequency)
        self.modulating_frequency = as_Hz(modulating_frequency)
        self.signal_delay = as_s(signal_delay)

    ##############################################

    def format_spice_parameters(self):
        # Fixme: to func?
        return ('AM(' +
                join_list((self.offset, self.amplitude, self.carrier_frequency,
                           self.modulating_frequency, self.signal_delay)) +
                ')')

####################################################################################################

class RandomMixin(SourceMixinAbc):

    r"""This class implements a Random Voltage source.

    The TRRANDOM option yields statistically distributed voltage values, derived from the ngspice
    random number generator. These values may be used in the transient simulation directly within a
    circuit, e.g. for generating a specific noise voltage, but especially they may be used in the
    control of behavioral sources (B, E, G sources, voltage controllable A sources, capacitors,
    inductors, or resistors) to simulate the circuit dependence on statistically varying device
    parameters. A Monte-Carlo simulation may thus be handled in a single simulation run.

    Spice Syntax::

        TRRANDOM( TYPE TS <TD <PARAM1 <PARAM2> > >)

    TYPE determines the random variates generated: 1 is uniformly distributed, 2 Gaussian, 3
    exponential, 4 Poisson. TS is the duration of an individual voltage value. TD is a time delay
    with 0 V output before the random voltage values start up. PARAM1 and PARAM2 depend on the type
    selected.

    +-------------+---------------+---------+-------------+---------+
    | Type        + Parameter 1   + Default + Parameter 2 + Default |
    +-------------+---------------+---------+-------------+---------+
    | uniform     + range         + 1       + offset      + 0       |
    +-------------+---------------+---------+-------------+---------+
    | gaussian    + standard dev. + 1       + mean        + 0       |
    +-------------+---------------+---------+-------------+---------+
    | exponential + mean          + 1       + offset      + 0       |
    +-------------+---------------+---------+-------------+---------+
    | poisson     + lambda        + 1       + offset      + 0       |
    +-------------+---------------+---------+-------------+---------+

    """

    ##############################################

    def __init__(self, random_type, duration=0, time_delay=0, parameter1=1, parameter2=0):
        # Fixme: random_type and parameters
        self.random_type = random_type
        self.duration = as_s(duration)
        self.time_delay = as_s(time_delay)
        self.parameter1 = parameter1
        self.parameter2 = parameter2

    ##############################################

    def format_spice_parameters(self):

        if self.random_type == 'uniform':
            random_type = 1
        elif self.random_type == 'exponential':
            random_type = 2
        elif self.random_type == 'gaussian':
            random_type = 3
        elif self.random_type == 'poisson':
            random_type = 4
        else:
            raise ValueError("Wrong random type {}".format(self.random_type))

        # Fixme: to func?
        return ('TRRANDOM(' +
                join_list((random_type, self.duration, self.time_delay,
                           self.parameter1, self.parameter2)) +
                ')')

####################################################################################################

class SinusoidalVoltageSource(VoltageSource, VoltageSourceMixinAbc, SinusoidalMixin):

    r"""This class implements a sinusoidal waveform voltage source.

    See :class:`SinusoidalMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
        SinusoidalMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = SinusoidalMixin.format_spice_parameters

####################################################################################################

class SinusoidalCurrentSource(CurrentSource, CurrentSourceMixinAbc, SinusoidalMixin):

    r"""This class implements a sinusoidal waveform current source.

    See :class:`SinusoidalMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
        SinusoidalMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = SinusoidalMixin.format_spice_parameters

####################################################################################################

class AcLine(SinusoidalVoltageSource):

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, rms_voltage=230, frequency=50):
        super().__init__(netlist, name, node_plus, node_minus,
                         amplitude=rms_to_amplitude(rms_voltage),
                         frequency=frequency)

####################################################################################################

class PulseVoltageSource(VoltageSource, VoltageSourceMixinAbc, PulseMixin):

    r"""This class implements a pulse waveform voltage source.

    See :class:`PulseMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
        PulseMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = PulseMixin.format_spice_parameters

####################################################################################################

class PulseCurrentSource(CurrentSource, CurrentSourceMixinAbc, PulseMixin):

    r"""This class implements a pulse waveform current source.

    See :class:`PulseMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
        PulseMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = PulseMixin.format_spice_parameters

####################################################################################################

class ExponentialVoltageSource(VoltageSource, VoltageSourceMixinAbc, ExponentialMixin):

    r"""This class implements a exponential waveform voltage source.

    See :class:`ExponentialMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
        ExponentialMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = ExponentialMixin.format_spice_parameters

####################################################################################################

class ExponentialCurrentSource(CurrentSource, CurrentSourceMixinAbc, ExponentialMixin):

    r"""This class implements a exponential waveform current source.

    See :class:`ExponentialMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
        ExponentialMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = ExponentialMixin.format_spice_parameters

####################################################################################################

class PieceWiseLinearVoltageSource(VoltageSource, VoltageSourceMixinAbc, PieceWiseLinearMixin):

    r"""This class implements a piece wise linear waveform voltage source.

    See :class:`PieceWiseLinearMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
        PieceWiseLinearMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = PieceWiseLinearMixin.format_spice_parameters

####################################################################################################

class PieceWiseLinearCurrentSource(CurrentSource, CurrentSourceMixinAbc, PieceWiseLinearMixin):

    r"""This class implements a piece wise linear waveform current source.

    See :class:`PieceWiseLinearMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
        PieceWiseLinearMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = PieceWiseLinearMixin.format_spice_parameters

####################################################################################################

class SingleFrequencyFMVoltageSource(VoltageSource, VoltageSourceMixinAbc, SingleFrequencyFMMixin):

    r"""This class implements a single frequency FM waveform voltage source.

    See :class:`SingleFrequencyFMMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):

        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
        SingleFrequencyFMMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = SingleFrequencyFMMixin.format_spice_parameters

####################################################################################################

class SingleFrequencyFMCurrentSource(CurrentSource, CurrentSourceMixinAbc, SingleFrequencyFMMixin):

    r"""This class implements a single frequency FM waveform current source.

    See :class:`SingleFrequencyFMMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
        SingleFrequencyFMMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = SingleFrequencyFMMixin.format_spice_parameters

####################################################################################################

class AmplitudeModulatedVoltageSource(VoltageSource, VoltageSourceMixinAbc, AmplitudeModulatedMixin):

    r"""This class implements a amplitude modulated waveform voltage source.

    See :class:`AmplitudeModulatedMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
        AmplitudeModulatedMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = AmplitudeModulatedMixin.format_spice_parameters

####################################################################################################

class AmplitudeModulatedCurrentSource(CurrentSource, CurrentSourceMixinAbc, AmplitudeModulatedMixin):

    r"""This class implements a amplitude modulated waveform current source.

    See :class:`AmplitudeModulatedMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
        AmplitudeModulatedMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = AmplitudeModulatedMixin.format_spice_parameters

####################################################################################################

class RandomVoltageSource(VoltageSource, VoltageSourceMixinAbc, RandomMixin):

    r"""This class implements a random waveform voltage source.

    See :class:`RandomMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        VoltageSource.__init__(self, netlist, name, node_plus, node_minus)
        RandomMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = RandomMixin.format_spice_parameters

####################################################################################################

class RandomCurrentSource(CurrentSource, CurrentSourceMixinAbc, RandomMixin):

    r"""This class implements a random waveform current source.

    See :class:`RandomMixin` for documentation.

    """

    ##############################################

    def __init__(self, netlist, name, node_plus, node_minus, *args, **kwargs):
        CurrentSource.__init__(self, netlist, name, node_plus, node_minus)
        RandomMixin.__init__(self, *args, **kwargs)

    ##############################################

    format_spice_parameters = RandomMixin.format_spice_parameters



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import os

####################################################################################################

def to_absolute_path(path):

    # Expand ~ . and Remove trailing '/'

    return os.path.abspath(os.path.expanduser(path))

####################################################################################################

def parent_directory_of(file_name, step=1):

    directory = file_name
    for i in range(step):
        directory = os.path.dirname(directory)
    return directory

####################################################################################################

def find(file_name, directories):

    if isinstance(directories, bytes):
        directories = (directories,)
    for directory in directories:
        for directory_path, sub_directories, file_names in os.walk(directory):
            if file_name in file_names:
                return os.path.join(directory_path, file_name)

    raise NameError("File %s not found in directories %s" % (file_name, str(directories)))



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2019 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import unittest

####################################################################################################

from PySpice.Spice.HighLevelElement import *
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

class TestHighLevelElement(unittest.TestCase):

    ##############################################

    def _test_spice_declaration(self, element, spice_declaration):
        self.assertEqual(str(element), spice_declaration)

    ##############################################

    def test(self):

        self._test_spice_declaration(
            PieceWiseLinearVoltageSource(
                Circuit(''),
                'pwl1', '1', '0',
                values=[(0, 0), (10@u_ms, 0), (11@u_ms, 5@u_V), (20@u_ms, 5@u_V)],
            ),
            'Vpwl1 1 0 PWL(0s 0V 10ms 0V 11ms 5V 20ms 5V)',
        )

        self._test_spice_declaration(
            PieceWiseLinearVoltageSource(
                Circuit(''),
                'pwl1', '1', '0',
                values=[(0, 0), (10@u_ms, 0), (11@u_ms, 5@u_V), (20@u_ms, 5@u_V)],
                repeat_time=12@u_ms, delay_time=34@u_ms,
            ),
            'Vpwl1 1 0 PWL(0s 0V 10ms 0V 11ms 5V 20ms 5V r=12ms td=34ms)',
        )

        self._test_spice_declaration(
            PieceWiseLinearVoltageSource(
                Circuit(''),
                'pwl1', '1', '0',
                values=[(0, 0), (10@u_ms, 0), (11@u_ms, 5@u_V), (20@u_ms, 5@u_V)],
                repeat_time=12@u_ms, delay_time=34@u_ms,
                dc=50@u_V,
            ),
            'Vpwl1 1 0 DC 50V PWL(0s 0V 10ms 0V 11ms 5V 20ms 5V r=12ms td=34ms)',
        )

####################################################################################################

if __name__ == '__main__':
    unittest.main()



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2020 jmgc / Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

"""This module implements a partial SPICE netlist parser.

See the :command:`cir2py` tool for an example of usage of the parser.

It would be difficult to implement a full parser for Ngspice since the syntax is mainly contextual.

SPICE is case insensitive.

"""

####################################################################################################

from collections import OrderedDict
import logging
import os
import regex

####################################################################################################

from .ElementParameter import FlagParameter
from .Netlist import ElementParameterMetaClass, Circuit, SubCircuit

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class ParseError(NameError):
    pass

####################################################################################################

class PrefixData:

    """This class represents a device prefix."""

    ##############################################

    def __init__(self, prefix, classes):

        self.prefix = prefix
        self.classes = classes

        number_of_positionals_min = 1000
        number_of_positionals_max = 0
        has_optionals = False
        for element_class in classes:
            number_of_positionals = element_class.number_of_positional_parameters
            number_of_positionals_min = min(number_of_positionals_min, number_of_positionals)
            number_of_positionals_max = max(number_of_positionals_max, number_of_positionals)
            has_optionals = max(has_optionals, bool(element_class.optional_parameters))

        self.number_of_positionals_min = number_of_positionals_min
        self.number_of_positionals_max = number_of_positionals_max
        self.has_optionals = has_optionals

        self.multi_devices = len(classes) > 1
        self.has_variable_number_of_pins = prefix in ('Q', 'X')  # NPinElement, Q has 3 to 4 pins
        if self.has_variable_number_of_pins:
            self.number_of_pins = None
        else:
            # Q and X are single
            self.number_of_pins = classes[0].number_of_pins

        self.has_flag = False
        for element_class in classes:
            for parameter in element_class.optional_parameters.values():
                if isinstance(parameter, FlagParameter):
                    self.has_flag = True

    ##############################################

    def __len__(self):
        return len(self.classes)

    ##############################################

    def __iter__(self):
        return iter(self.classes)

    ##############################################

    @property
    def single(self):
        if not self.multi_devices:
            return self.classes[0]
        else:
            raise NameError()

####################################################################################################

_prefix_cache = {}
for prefix, classes in ElementParameterMetaClass._classes.items():
    prefix_data = PrefixData(prefix, classes)
    _prefix_cache[prefix] = prefix_data
    _prefix_cache[prefix.lower()] = prefix_data

# for prefix_data in sorted(_prefix_cache.values(), key=lambda x: len(x)):
#     print(prefix_data.prefix,
#           len(prefix_data),
#           prefix_data.number_of_positionals_min, prefix_data.number_of_positionals_max,
#           prefix_data.has_optionals)

# Single:
# B 0 True
# D 1 True
# F 2 False
# G 1 False
# H 2 False
# I 1 False
# J 1 True
# K 3 False
# M 1 True
# S 2 False
# V 1 False
# W 3 False
# Z 1 True

# Two:
# E 0 1 False
# L 1 2 True

# Three:
# C 1 2 True
# R 1 2 True

# NPinElement:
# Q 1 1 True
# X 1 1 False

####################################################################################################

class Statement:

    """This base class implements a statement, in fact a line in a Spice netlist."""

    ##############################################

    def __init__(self, line, statement=None):
        self._line = line
        if statement is not None:
            self._line.lower_case_statement(statement)

    ##############################################

    def __repr__(self):
        return '{} {}'.format(self.__class__.__name__, repr(self._line))

    ##############################################

    def value_to_python(self, x):
        if x:
            if str(x)[0].isdigit():
                return str(x)
            else:
                return "'{}'".format(x)
        else:
            return ''

    ##############################################

    def values_to_python(self, values):
        return [self.value_to_python(x) for x in values]

    ##############################################

    def kwargs_to_python(self, kwargs):
        return ['{}={}'.format(key, self.value_to_python(value))
                for key, value in kwargs.items()]

    ##############################################

    def join_args(self, args):
        return ', '.join(args)

####################################################################################################

class Comment(Statement):
    pass

####################################################################################################

class Title(Statement):

    """This class implements a title definition."""

    ##############################################

    def __init__(self, line):
        super().__init__(line, statement='title')
        self._title = self._line.right_of('.title')

    ##############################################

    def __str__(self):
        return self._title

    ##############################################

    def __repr__(self):
        return 'Title {}'.format(self._title)

####################################################################################################

class Include(Statement):

    """This class implements a include definition."""

    ##############################################

    def __init__(self, line):
        super().__init__(line, statement='include')
        self._include = self._line.right_of('.include')

    ##############################################

    def __str__(self):
        return self._include

    ##############################################

    def __repr__(self):
        return 'Include {}'.format(self._include)

    ##############################################

    def to_python(self, netlist_name):
        return '{}.include({})'.format(netlist_name, self._include) + os.linesep

####################################################################################################

class Model(Statement):

    """This class implements a model definition.

    Spice syntax::

        .model mname type(pname1=pval1 pname2=pval2 ... )

    """

    ##############################################

    def __init__(self, line):
        super().__init__(line, statement='model')

        base, self._parameters = line.split_keyword('.model')
        self._name, self._model_type = base
        self._name = self._name.lower()

    ##############################################

    @property
    def name(self):
        """Name of the model"""
        return self._name

    ##############################################

    def __repr__(self):
        return 'Model {} {} {}'.format(self._name, self._model_type, self._parameters)

    ##############################################

    def to_python(self, netlist_name):
        args = self.values_to_python((self._name, self._model_type))
        kwargs = self.kwargs_to_python(self._parameters)
        return '{}.model({})'.format(netlist_name, self.join_args(args + kwargs)) + os.linesep

    ##############################################

    def build(self, circuit):
        return circuit.model(self._name, self._model_type, **self._parameters)

####################################################################################################

class Parameter(Statement):

    """This class implements a parameter definition.

    Spice syntax::

        .param name=expr

    """

    ##############################################

    def __init__(self, line):
        super().__init__(line, statement='param')

        text = line.right_of('.param').strip().lower()  # Fixme: lower ???
        idx = text.find('=')
        self._name = text[:idx].strip()
        self._value = text[idx + 1:].strip()

    ##############################################

    @property
    def name(self):
        """Name of the model"""
        return self._name

    ##############################################

    def __repr__(self):
        return 'Param {}={}'.format(self._name, self._value)

    ##############################################

    def to_python(self, netlist_name):
        args = self.values_to_python((self._name, self._value))
        # Fixme: linesep here ???
        return '{}.param({})'.format(netlist_name, self.join_args(args)) + os.linesep

    ##############################################

    def build(self, circuit):
        circuit.parameter(self._name, self._value)

####################################################################################################

# Review: HERE

class CircuitStatement(Statement):

    # Review: jmgc

    """This class implements a circuit definition.

    Spice syntax::

        Title ...

    """

    ##############################################

    def __init__(self, title):

        super().__init__(title, statement='title')

        # Review: Title
        title_statement = '.title '
        self._title = str(title)
        if self._title.startswith(title_statement):
            self._title = self._title[len(title_statement):]

        self._statements = []
        self._subcircuits = []
        self._models = []
        self._required_subcircuits = set()
        self._required_models = set()
        self._params = []

    ##############################################

    @property
    def title(self):
        """Title of the circuit."""
        return self._title

    @property
    def name(self):
        """Name of the circuit."""
        return self._title

    @property
    def models(self):
        """Models of the circuit."""
        return self._models

    @property
    def subcircuits(self):
        """Subcircuits of the circuit."""
        return self._subcircuits

    @property
    def params(self):
        """Parameters of the circuit."""
        return self._params

    ##############################################

    def __repr__(self):
        text = 'Circuit {}'.format(self._title) + os.linesep
        text += os.linesep.join([repr(model) for model in self._models]) + os.linesep
        text += os.linesep.join([repr(subcircuit) for subcircuit in self._subcircuits]) + os.linesep
        text += os.linesep.join(['  ' + repr(statement) for statement in self._statements])
        return text

    ##############################################

    def __iter__(self):
        """Return an iterator on the statements."""
        return iter(self._models + self._subcircuits + self._statements)

    ##############################################

    def append(self, statement):
        """Append a statement to the statement's list."""
        self._statements.append(statement)

    ##############################################

    def append_model(self, statement):
        """Append a model to the statement's list."""
        self._models.append(statement)

    ##############################################

    def append_param(self, statement):
        """Append a param to the statement's list."""
        self._params.append(statement)

    ##############################################

    def append_subcircuit(self, statement):
        """Append a subcircuit to the statement's list."""
        self._subcircuits.append(statement)

    ##############################################

    def to_python(self, ground=0):
        subcircuit_name = 'subcircuit_' + self._name
        args = self.values_to_python([subcircuit_name] + self._nodes)
        source_code = ''
        source_code += '{} = SubCircuit({})'.format(subcircuit_name, self.join_args(args)) + os.linesep
        source_code += SpiceParser.netlist_to_python(subcircuit_name, self, ground)
        return source_code

    ##############################################

    def build(self, ground=0):
        circuit = Circuit(self._title)
        for statement in self._params:
            statement.build(circuit)
        for statement in self._models:
            model = statement.build(circuit)
        for statement in self._subcircuits:
            subckt = statement.build(ground)  # Fixme: ok ???
            circuit.subcircuit(subckt)
        for statement in self._statements:
            if isinstance(statement, Element):
                statement.build(circuit, ground)
        return circuit

####################################################################################################

class SubCircuitStatement(Statement):

    """This class implements a sub-circuit definition.

    Spice syntax::

        .SUBCKT name node1 ... param1=value1 ...

    """

    ##############################################

    def __init__(self, line):

        super().__init__(line, statement='subckt')

        # Fixme
        parameters, dict_parameters = self._line.split_keyword('.subckt')
        # Review: syntax ???
        if parameters[-1].lower() == 'params:':
            parameters = parameters[:-1]
        self._name, self._nodes = parameters[0], parameters[1:]
        self._name = self._name.lower()
        self._parameters = dict_parameters

        self._statements = []
        self._subcircuits = []
        self._models = []
        self._required_subcircuits = set()
        self._required_models = set()
        self._params = []

    ##############################################

    @property
    def name(self):
        """Name of the sub-circuit."""
        return self._name

    @property
    def nodes(self):
        """Nodes of the sub-circuit."""
        return self._nodes

    @property
    def models(self):
        """Models of the sub-circuit."""
        return self._models

    @property
    def params(self):
        """Params of the sub-circuit."""
        return self._params

    @property
    def subcircuits(self):
        """Subcircuits of the sub-circuit."""
        return self._subcircuits

    ##############################################

    def __repr__(self):
        if self._parameters:
            text = 'SubCircuit {} {} Parameters: {}'.format(self._name, self._nodes, self._parameters) + os.linesep
        else:
            text = 'SubCircuit {} {}'.format(self._name, self._nodes) + os.linesep
        text += os.linesep.join([repr(model) for model in self._models]) + os.linesep
        text += os.linesep.join([repr(subcircuit) for subcircuit in self._subcircuits]) + os.linesep
        text += os.linesep.join(['  ' + repr(statement) for statement in self._statements])
        return text

    ##############################################

    def __iter__(self):
        """Return an iterator on the statements."""
        return iter(self._models + self._subcircuits + self._statements)

    ##############################################

    def append(self, statement):
        """Append a statement to the statement's list."""
        self._statements.append(statement)

    ##############################################

    def append_model(self, statement):
        """Append a model to the statement's list."""
        self._models.append(statement)

    ##############################################

    def append_param(self, statement):
        """Append a param to the statement's list."""
        self._params.append(statement)

    ##############################################

    def append_subcircuit(self, statement):
        """Append a model to the statement's list."""
        self._subcircuits.append(statement)

    ##############################################

    def to_python(self, ground=0):
        subcircuit_name = 'subcircuit_' + self._name
        args = self.values_to_python([subcircuit_name] + self._nodes)
        source_code = ''
        source_code += '{} = SubCircuit({})'.format(subcircuit_name, self.join_args(args)) + os.linesep
        source_code += SpiceParser.netlist_to_python(subcircuit_name, self, ground)
        return source_code

    ##############################################

    def build(self, ground=0, parent=None):
        subcircuit = SubCircuit(self._name, *self._nodes, **self._parameters)
        subcircuit.parent = parent
        for statement in self._params:
            statement.build(subcircuit)
        for statement in self._models:
            model = statement.build(subcircuit)
        for statement in self._subcircuits:
            subckt = statement.build(ground, parent=subcircuit)  # Fixme: ok ???
            subcircuit.subcircuit(subckt)
        for statement in self._statements:
            if isinstance(statement, Element):
                statement.build(subcircuit, ground)
        return subcircuit

####################################################################################################

class Element(Statement):

    """This class implements an element definition.

    "{ expression }" are allowed in device line.

    """

    _logger = _module_logger.getChild('Element')

    ##############################################

    def __init__(self, line):

        super().__init__(line)

        line_str = str(line)
        # self._logger.debug(os.linesep + line_str)

        # Retrieve device prefix
        prefix = line_str[0]
        if prefix.isalpha():
            self._prefix = prefix
        else:
            raise ParseError("Not an element prefix: " + prefix)
        prefix_data = _prefix_cache[self._prefix]

        # Retrieve device name
        args, kwargs = line.split_element(prefix)
        self._name = args.pop(0)

        self._nodes = []
        self._parameters = []
        self._dict_parameters = {}

        # Read nodes
        if not prefix_data.has_variable_number_of_pins:
            number_of_pins = prefix_data.number_of_pins
            if number_of_pins:
                self._nodes = args[:number_of_pins]
                args = args[number_of_pins:]
        else:  # Q or X
            if prefix_data.prefix == 'Q':
                self._nodes = args[:3]
                args = args[3:]
                # Fixme: optional node
            else:  # X
                if args[-1].lower() == 'params:':
                    args.pop()
                self._parameters.append(args.pop())
                self._nodes = args
                args = []

        # Read positionals
        number_of_positionals = prefix_data.number_of_positionals_min
        if number_of_positionals and (len(args) > 0) and (prefix_data.prefix != 'X'):  # model is optional
            self._parameters = args[:number_of_positionals]
            args = args[number_of_positionals:]
        if prefix_data.multi_devices and (len(args) > 0):
            remaining = args
            args = []
            self._parameters.extend(remaining)

        if prefix_data.prefix in ('V', 'I') and (len(args) > 0):
            # merge remaining
            self._parameters[-1] += " " + " ".join(args)
            self._dict_parameters = kwargs

        # Read optionals
        if (prefix_data.has_optionals or (prefix_data.prefix == 'X')) and (len(kwargs) > 0):
            for key in kwargs:
                self._dict_parameters[key] = kwargs[key]

        if prefix_data.multi_devices:
            for element_class in prefix_data:
                if len(self._parameters) == element_class.number_of_positional_parameters:
                    break
        else:
            element_class = prefix_data.single
        self.factory = element_class

        # Move positionals passed as kwarg
        to_delete = []
        for parameter in element_class.positional_parameters.values():
            if parameter.key_parameter:
                idx = parameter.position
                if idx < len(self._parameters):
                    self._dict_parameters[parameter.attribute_name] = self._parameters[idx]
                    to_delete.append(idx - len(to_delete))
        for idx in to_delete:
            self._parameters.pop(idx)

        # self._logger.debug(os.linesep + self.__repr__())

    ##############################################

    @property
    def name(self):
        """Name of the element"""
        return self._name

    ##############################################

    def __repr__(self):
        return 'Element {0._prefix} {0._name} {0._nodes} {0._parameters} {0._dict_parameters}'.format(self)

    ##############################################

    def translate_ground_node(self, ground):
        nodes = []
        for node in self._nodes:
            if str(node) == str(ground):
                node = 0
            nodes.append(node)
        return nodes

    ##############################################

    def to_python(self, netlist_name, ground=0):

        nodes = self.translate_ground_node(ground)
        args = [self._name]
        if self._prefix != 'X':
            args += nodes + self._parameters
        else:  # != Spice
            args += self._parameters + nodes
        args = self.values_to_python(args)
        kwargs = self.kwargs_to_python(self._dict_parameters)
        return '{}.{}({})'.format(netlist_name, self._prefix, self.join_args(args + kwargs)) + os.linesep

    ##############################################

    def _check_params(self, elements=1):
        params = []
        for param in self._parameters:
            values = param.replace(',', ' ')
            if values[0] == '(' and values[-1] == ')':
                values = values[1: -1].split()
                if len(values) > elements:
                    raise IndexError('Incorrect number of elements for (%r): %s' % (self, param))
                params.extend(values)
            else:
                params.extend(values.split())
        self._parameters = params

    ##############################################

    def _voltage_controlled_nodes(self, poly_arg):
        result = ['v(%s,%s)' % nodes
                  for nodes in zip(self._parameters[:(2 * poly_arg):2],
                                   self._parameters[1:(2 * poly_arg):2])]
        result += self._parameters[2 * poly_arg:]
        return ' '.join(result)

    ##############################################

    def _current_controlled_nodes(self, poly_arg):
        result = ['i(%s)' % node
                  for node in self._parameters[:poly_arg]]
        result += self._parameters[poly_arg:]
        return ' '.join(result)

    ##############################################

    def _manage_controlled_sources(self, nodes):
        try:
            idx = self._nodes.index('POLY')
            if idx == 2:
                poly_arg = self._nodes[3]
                if poly_arg[0] == '(' and poly_arg[-1] == ')':
                    poly_arg = poly_arg[1:-1]
                try:
                    poly_arg = int(poly_arg)
                except TypeError as te:
                    raise TypeError('Not valid poly argument: %s' % poly_arg, te)
                self._nodes = self._nodes[:2]
                nodes = nodes[:2]
                if self._prefix in 'EG':
                    self._check_params(2)
                    values = self._voltage_controlled_nodes(poly_arg)
                    if self._prefix == 'E':
                        key = 'v'
                    else:
                        key = 'i'
                else:
                    self._check_params(1)
                    values = self._current_controlled_nodes(poly_arg)
                    if self._prefix == 'F':
                        key = 'v'
                    else:
                        key = 'i'
                poly_str = '{ POLY (%d) %s }' % (poly_arg, values)

                self._dict_parameters[key] = poly_str
                self._parameters.clear()
                self._name = self._prefix + self._name
                self._prefix = 'B'
                prefix_data = _prefix_cache[self._prefix]
                self.factory = prefix_data.single
                return nodes
            raise IndexError('Incorrect position of POLY: %r' % self)
        except ValueError:
            pass
        _correction = []
        correction = []
        for _node, node in zip(self._nodes, nodes):
            _values = _node.replace(',', ' ')
            try:
                values = node.replace(',', ' ')
            except AttributeError:
                values = str(node)
            if _values[0] == '(' and _values[-1] == ')':
                _values = _values[1: -1]
            if values[0] == '(' and values[-1] == ')':
                values = values[1: -1]
            _correction.extend(_values.split())
            correction.extend(values.split())
        self._parameters = correction[len(self._nodes):] + self._parameters
        self._nodes = _correction[:len(self._nodes)]
        parameters = self._parameters
        correction = correction[:len(self._nodes)]
        if self._prefix in 'EG':
            if len(correction) + len(parameters) == 5:
                parameters = correction[2:] + parameters
                self._nodes = _correction[:2]
                value = '{v(%s, %s) * %s}' % tuple(parameters)
                if self._prefix == 'E':
                    key = 'v'
                else:
                    key = 'i'
                self._dict_parameters[key] = value
                self._parameters.clear()
                self._name = self._prefix + self._name
                self._prefix = 'B'
                prefix_data = _prefix_cache[self._prefix]
                self.factory = prefix_data.single
        else:
            if len(correction) + len(parameters) == 4:
                parameters = correction[2:] + parameters
                self._nodes = _correction[:2]
                value = '{i(%s) * %s}' % tuple(parameters)
                if self._prefix == 'F':
                    key = 'v'
                else:
                    key = 'i'
                self._dict_parameters[key] = value
                self._parameters.clear()
                self._name = self._prefix + self._name
                self._prefix = 'B'
                prefix_data = _prefix_cache[self._prefix]
                self.factory = prefix_data.single
        return correction[:len(self._nodes)]

    ##############################################

    def build(self, circuit, ground=0):

        nodes = self.translate_ground_node(ground)
        if self._prefix != 'X':
            if self._prefix in ('EFGH'):
                nodes = self._manage_controlled_sources(nodes)
            args = nodes + self._parameters
        else:  # != Spice
            args = self._parameters + nodes
        factory = getattr(circuit, self.factory.__alias__)
        kwargs = self._dict_parameters
        message = ' '.join([str(x) for x in (self._prefix, self._name, args,
                                             self._dict_parameters)])
        self._logger.debug(message)
        return factory(self._name, *args, **kwargs)


####################################################################################################

class Line:

    """This class implements a line in the netlist."""

    _logger = _module_logger.getChild('Line')

    ##############################################

    def __init__(self, line, line_range, end_of_line_comment):

        self._end_of_line_comment = end_of_line_comment

        text, comment, self._is_comment = self._split_comment(line)

        self._text = text
        self._comment = comment
        self._line_range = line_range

    ##############################################

    def __repr__(self):
        return '{0._line_range}: {0._text} // {0._comment}'.format(self)

    ##############################################

    def __str__(self):
        return self._text

    ##############################################

    @property
    def comment(self):
        return self._comment

    @property
    def is_comment(self):
        return self._is_comment

    ##############################################

    def _split_comment(self, line):

        line = str(line)

        if line.startswith('*'):
            is_comment = True
            text = ''
            comment = line[1:].strip()
        else:
            is_comment = False
            # remove end of line comment
            location = -1
            for marker in self._end_of_line_comment:
                _location = line.find(marker)
                if _location != -1:
                    if location == -1:
                        location = _location
                    else:
                        location = min(_location, location)
            if location != -1:
                text = line[:location].strip()
                comment = line[location:].strip()
            else:
                text = line
                comment = ''

        return text, comment, is_comment

    ##############################################

    def append(self, line):

        text, comment, is_comment = self._split_comment(line)

        if text:
            if not self._text.endswith(' ') or text.startswith(' '):
                self._text += ' '
            self._text += text
        if comment:
            self._comment += ' // ' + comment

        _slice = self._line_range
        self._line_range = slice(_slice.start, _slice.stop + 1)

    ##############################################

    def lower_case_statement(self, statement):

        """Lower case the statement"""

        # statement without . prefix

        if self._text:
            lower_statement = statement.lower()
            _slice = slice(1, len(statement) + 1)
            _statement = self._text[_slice]
            if _statement.lower() == lower_statement:
                self._text = '.' + lower_statement + self._text[_slice.stop:]

    ##############################################

    def right_of(self, text):
        return self._text[len(text):].strip()

    ##############################################

    def read_words(self, start_location, number_of_words):

        """Read a fixed number of words separated by space."""

        words = []
        stop_location = None

        line_str = self._text
        number_of_words_read = 0
        while number_of_words_read < number_of_words:  # and start_location < len(line_str)
            if line_str[start_location] == '{':
                stop_location = line_str.find('}', start_location)
                if stop_location > start_location:
                    stop_location += 1
            else:
                stop_location = line_str.find(' ', start_location)
            if stop_location == -1:
                stop_location = None  # read until end
            word = line_str[start_location:stop_location].strip()
            if word:
                number_of_words_read += 1
                words.append(word)
            if stop_location is None:  # we should stop
                if number_of_words_read != number_of_words:
                    template = 'Bad element line, looking for word {}/{}:' + os.linesep
                    message = (template.format(number_of_words_read, number_of_words) +
                               line_str + os.linesep +
                               ' ' * start_location + '^')
                    self._logger.warning(message)
                    raise ParseError(message)
            else:
                if start_location < stop_location:
                    start_location = stop_location
                else:  # we have read a space
                    start_location += 1

        return words, stop_location

    ##############################################

    def split_words(self, start_location, until=None):

        stop_location = None

        line_str = self._text
        if until is not None:
            location = line_str.find(until, start_location)
            if location != -1:
                stop_location = location
                location = line_str.rfind(' ', start_location, stop_location)
                if location != -1:
                    stop_location = location
                else:
                    raise NameError('Bad element line, missing key? ' + line_str)

        line_str = line_str[start_location:stop_location]
        words = [x for x in line_str.split(' ') if x]
        result = []
        expression = 0
        begin_idx = 0
        for idx, word in enumerate(words):
            if expression == 0:
                begin_idx = idx
            expression += word.count('{') - word.count('}')
            if expression == 0:
                if begin_idx < idx:
                    result.append(' '.join(words[begin_idx:idx + 1]))
                else:
                    result.append(word)
        return result, stop_location

    ##############################################

    @staticmethod
    def get_kwarg(text):

        dict_parameters = {}

        parts = []
        for part in text.split():
            if '=' in part and part != '=':
                left, right = [x for x in part.split('=')]
                parts.append(left)
                parts.append('=')
                if right:
                    parts.append(right)
            else:
                parts.append(part)

        i = 0
        i_stop = len(parts)
        while i < i_stop:
            if i + 1 < i_stop and parts[i + 1] == '=':
                key, value = parts[i], parts[i + 2]
                dict_parameters[key] = value
                i += 3
            else:
                raise ParseError("Bad kwarg: {}".format(text))

        return dict_parameters

    ##############################################

    @staticmethod
    def _partition(text):
        parts = []
        values = text.replace(',', ' ')
        for part in values.split():
            if '=' in part and part != '=':
                left, right = [x for x in part.split('=')]
                parts.append(left)
                parts.append('=')
                if right:
                    parts.append(right)
            else:
                parts.append(part)
        return parts

    ##############################################

    @staticmethod
    def _partition_parentheses(text):
        p = regex.compile(r'\(([^\(\)]|(?R))*?\)')
        parts = []
        previous_start = 0
        for m in regex.finditer(p, text):
            parts.extend(Line._partition(text[previous_start:m.start()]))
            parts.append(m.group())
            previous_start = m.end()
        parts.extend(Line._partition(text[previous_start:]))
        return parts

    ##############################################

    @staticmethod
    def _partition_braces(text):
        p = regex.compile(r'\{([^\{\}]|(?R))*?\}')
        parts = []
        previous_start = 0
        for m in regex.finditer(p, text):
            parts.extend(Line._partition_parentheses(text[previous_start:m.start()]))
            parts.append(m.group())
            previous_start = m.end()
        parts.extend(Line._partition_parentheses(text[previous_start:]))
        return parts

    ##############################################

    @staticmethod
    def _check_parameters(parts):
        parameters = []
        dict_parameters = {}

        i = 0
        i_stop = len(parts)
        while i < i_stop:
            if i + 1 < i_stop and parts[i + 1] == '=':
                key, value = parts[i], parts[i + 2]
                dict_parameters[key] = value
                i += 3
            else:
                parameters.append(parts[i])
                i += 1

        return parameters, dict_parameters

    ##############################################

    def split_keyword(self, keyword):

        """Split the line according to the following pattern::

            keyword parameter1 parameter2 ( key1=value1 key2=value2 )

        Return the list of parameters and the dictionary.
        The parenthesis can be omitted.

        """

        text = self.right_of(keyword)

        p = regex.compile(r'\(([^\(\)]|(?R))*?\)')
        b = regex.compile(r'\{([^\{\}]|(?R))*?\}')
        parts = []

        mp = regex.search(p, text)
        mb = regex.search(b, text)
        if mb is not None:
            if mp is not None:
                if (mb.start() > mp.start()) and (mb.end() < mp.end()):
                    parts.extend(Line._partition(text[:mp.start()]))
                    parts.extend(Line._partition_braces(mp.group()[1:-1]))
                elif (mb.start() < mp.start()) and (mb.end() > mp.end()):
                    parts.extend(Line._partition_braces(text))
                else:
                    raise ValueError("Incorrect format {}".format(text))
            else:
                parts.extend(Line._partition_braces(text))
        else:
            if mp is not None:
                parts.extend(Line._partition(text[:mp.start()]))
                parts.extend(Line._partition(mp.group()[1:-1]))
            else:
                parts.extend(Line._partition(text))
        return Line._check_parameters(parts)

    ##############################################

    def split_element(self, prefix):

        """Split the line according to the following pattern::

            keyword parameter1 parameter2 ... key1=value1 key2=value2 ...

        Return the list of parameters and the dictionary.

        """

        # Fixme: cf. get_kwarg

        parameters = []
        dict_parameters = {}

        text = self.right_of(prefix)

        parts = Line._partition_braces(text)

        return Line._check_parameters(parts)

####################################################################################################

class SpiceParser:

    """This class parse a Spice netlist file and build a syntax tree.

    Public Attributes:

      :attr:`circuit`

      :attr:`models`

      :attr:`subcircuits`

    """

    _logger = _module_logger.getChild('SpiceParser')

    ##############################################

    def __init__(self, path=None, source=None, end_of_line_comment=('$', '//', ';')):

        # Fixme: empty source

        if path is not None:
            with open(str(path), 'r') as fh:
                raw_lines = fh.readlines()  # Fixme: cf. jmgc
        elif source is not None:
            raw_lines = source.split(os.linesep)
        else:
            raise ValueError

        self._end_of_line_comment = end_of_line_comment

        lines = self._merge_lines(raw_lines)
        self._title = None
        self._statements = self._parse(lines)

    ##############################################

    def _merge_lines(self, raw_lines):

        """Merge broken lines and return a new list of lines.

        A line starting with "+" continues the preceding line.

        """

        lines = []
        current_line = None
        for line_index, line_string in enumerate(raw_lines):
            if line_string.startswith('+'):
                current_line.append(line_string[1:].strip('\r\n'))
            else:
                line_string = line_string.strip(' \t\r\n')
                if line_string:
                    _slice = slice(line_index, line_index + 1)
                    line = Line(line_string, _slice, self._end_of_line_comment)
                    lines.append(line)
                    # handle case with comment before line continuation
                    if not line_string.startswith('*'):
                        current_line = line

        return lines

    ##############################################

    @staticmethod
    def _check_models(circuit, available_models=set()):
        p_available_models = available_models.copy()
        p_available_models.update([model.name for model in circuit._models])
        for subcircuit in circuit._subcircuits:
            SpiceParser._check_models(subcircuit, p_available_models)
        for model in circuit._required_models:
            if model not in p_available_models:
                raise ValueError("model (%s) not available in (%s)" % (model, circuit.name))

    ##############################################

    @staticmethod
    def _sort_subcircuits(circuit, available_subcircuits=set()):
        p_available_subcircuits = available_subcircuits.copy()
        names = [subcircuit.name for subcircuit in circuit._subcircuits]
        p_available_subcircuits.update(names)
        dependencies = dict()
        for subcircuit in circuit._subcircuits:
            required = SpiceParser._sort_subcircuits(subcircuit, p_available_subcircuits)
            dependencies[subcircuit] = required
        for subcircuit in circuit._required_subcircuits:
            if subcircuit not in p_available_subcircuits:
                raise ValueError("subcircuit (%s) not available in (%s)" % (subcircuit, circuit.name))
        items = sorted(dependencies.items(), key=lambda item: len(item[1]))
        result = list()
        result_names = list()
        previous = len(items) + 1
        while 0 < len(items) < previous:
            previous = len(items)
            remove = list()
            for item in items:
                subckt, depends = item
                for name in depends:
                    if name not in result_names:
                        break
                else:
                    result.append(subckt)
                    result_names.append(subckt.name)
                    remove.append(item)
            for item in remove:
                items.remove(item)
        if len(items) > 0:
            raise ValueError("Crossed dependencies (%s)" % [(key.name, value) for key, value in items])
        circuit._subcircuits = result
        return circuit._required_subcircuits - set(names)

    ##############################################

    def _parse(self, lines):

        """Parse the lines and return a list of statements."""

        # The first line in the input file must be the title, which is the only comment line that does
        # not need any special character in the first place.
        #
        # The last line must be .end

        if len(lines) <= 1:
            raise NameError('Netlist is empty')
        # if lines[-1] != '.end':
        #     raise NameError('".end" is expected at the end of the netlist')

        circuit = CircuitStatement(lines[0])
        stack = []
        scope = circuit
        for line in lines[1:]:
            # print('>', repr(line))
            text = str(line)
            lower_case_text = text.lower()  # !
            if line.is_comment:
                scope.append(Comment(line))
            elif lower_case_text.startswith('.'):
                lower_case_text = lower_case_text[1:]
                if lower_case_text.startswith('subckt'):
                    stack.append(scope)
                    scope = SubCircuitStatement(line)
                elif lower_case_text.startswith('ends'):
                    parent = stack.pop()
                    parent.append_subcircuit(scope)
                    scope = parent
                elif lower_case_text.startswith('title'):
                    # override fist line
                    self._title = Title(line)
                    scope.append(self._title)
                elif lower_case_text.startswith('end'):
                    pass
                elif lower_case_text.startswith('model'):
                    model = Model(line)
                    scope.append_model(model)
                elif lower_case_text.startswith('include'):
                    include = Include(line)
                    scope.append(include)
                elif lower_case_text.startswith('param'):
                    param = Parameter(line)
                    scope.append_param(param)
                else:
                    # options param ...
                    # .global
                    # .lib filename libname
                    # .func .csparam .temp .if
                    # { expr } are allowed in .model lines and in device lines.
                    self._logger.warn('Parser ignored: {}'.format(line))
            else:
                try:
                    element = Element(line)
                    scope.append(element)
                    if hasattr(element, '_prefix') and (element._prefix == "X"):
                        name = element._parameters[0].lower()
                        scope._required_subcircuits.add(name)
                    elif hasattr(element, '_dict_parameters') and 'model' in element._dict_parameters:
                        name = element._dict_parameters['model'].lower()
                        scope._required_models.add(name)
                except ParseError:
                    pass
        SpiceParser._check_models(circuit)
        SpiceParser._sort_subcircuits(circuit)
        return circuit

    ##############################################

    @property
    def circuit(self):
        """Circuit statements."""
        return self._statements

    @property
    def models(self):
        """Models of the sub-circuit."""
        return self._statements.models

    @property
    def subcircuits(self):
        """Subcircuits of the sub-circuit."""
        return self._statements.subcircuits

    ##############################################

    def is_only_subcircuit(self):
        return bool(not self.circuit and self.subcircuits)

    ##############################################

    def is_only_model(self):
        return bool(not self.circuit and not self.subcircuits and self.models)

    ##############################################

    @staticmethod
    def _build_circuit(circuit, statements, ground):

        for statement in statements:
            if isinstance(statement, Include):
                circuit.include(str(statement))

        for statement in statements:
            if isinstance(statement, Element):
                statement.build(circuit, ground)
            elif isinstance(statement, Model):
                statement.build(circuit)
            elif isinstance(statement, SubCircuit):
                subcircuit = statement.build(ground)  # Fixme: ok ???
                circuit.subcircuit(subcircuit)

    ##############################################

    def build_circuit(self, ground=0):
        """Build a :class:`Circuit` instance.

        Use the *ground* parameter to specify the node which must be translated to 0 (SPICE ground node).

        """
        # circuit = Circuit(str(self._title))
        circuit = self.circuit.build(str(ground))
        return circuit

    ##############################################

    @staticmethod
    def netlist_to_python(netlist_name, statements, ground=0):

        source_code = ''
        for statement in statements:
            if isinstance(statement, Element):
                source_code += statement.to_python(netlist_name, ground)
            elif isinstance(statement, Include):
                pass
            elif isinstance(statement, Model):
                source_code += statement.to_python(netlist_name)
            elif isinstance(statement, SubCircuitStatement):
                source_code += statement.to_python(netlist_name)
            elif isinstance(statement, Include):
                source_code += statement.to_python(netlist_name)
        return source_code

    ##############################################

    def to_python_code(self, ground=0):

        ground = str(ground)

        source_code = ''

        if self.circuit:
            source_code += "circuit = Circuit('{}')".format(self._title) + os.linesep
        source_code += self.netlist_to_python('circuit', self._statements, ground)

        return source_code



#r# =====================================
#r#  Relay drived by a bipolar transistor
#r# =====================================

#r# This example shows the simulation of ...

####################################################################################################

import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Probe.Plot import plot
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

#?# #cm# relay.m4

period = 50@u_ms
pulse_width = period / 2

circuit = Circuit('Relay')

# circuit.V('digital', 'Vdigital', circuit.gnd, 5@u_V)
circuit.PulseVoltageSource('clock', 'clock', circuit.gnd, 0@u_V, 5@u_V, pulse_width, period, rise_time=5@u_ms, fall_time=5@u_ms)
circuit.R('base', 'clock', 'base', 100@u_Ω)
circuit.BJT(1, 'collector', 'base', circuit.gnd, model='bjt') # Q is mapped to BJT !
circuit.model('bjt', 'npn', bf=80, cjc=pico(5), rb=100)
circuit.V('analog', 'VccAnalog', circuit.gnd, 8@u_V)
circuit.R('relay', 'VccAnalog', 1, 50@u_Ω)
circuit.L('relay', 1, 'collector', 100@u_mH)
circuit.include(spice_library['1N5822']) # Schottky diode
diode = circuit.X('D', '1N5822', 'collector', 'VccAnalog')
# Fixme: subcircuit node
# diode.minus.add_current_probe(circuit)

####################################################################################################

figure, ax = plt.subplots(figsize=(20, 10))

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.transient(step_time=period/1000, end_time=period*1.1)

ax.set_title('')
ax.set_xlabel('Time [s]')
ax.set_ylabel('Voltage [V]')
ax.grid()
ax.plot(analysis.base)
ax.plot(analysis.collector)
# Fixme: current probe
# ax.plot((analysis['1'] - analysis.collector)/circuit.Rrelay.resistance)
ax.plot(analysis['1'] - analysis.collector)
ax.legend(('Vbase', 'Vcollector'), loc=(.05,.1))

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'relay.png')



####################################################################################################
#
# PySpice - A Spice package for Python
# Copyright (C) 2019 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from pathlib import Path
import re
import shutil

from invoke import task

####################################################################################################

STANDARD_PACKAGES = (
    'argparse',
    'atexit',
    'datetime',
    'glob',
    'hashlib',
    'importlib',
    'json',
    'logging',
    'operator',
    'os',
    'pathlib',
    'shutil',
    'signal',
    'stat',
    'subprocess',
    'sys',
    'time',
    'traceback',
)

@task()
def show_import(ctx):
    package = ctx.Package
    with ctx.cd(package):
        result = ctx.run("grep -r -h -E  '^(import|from) [a-zA-Z]' . | sort | uniq", hide='out')
    imports = set()
    for line in result.stdout.split('\n'):
        if line.startswith('from'):
            position = line.find('import')
            line = line[5:position]
        elif line.startswith('import'):
            line = line[7:]
        # print('|{}|'.format(line))
        position = line.find('.')
        if position != -1:
            line = line[:position]
        line = line.strip()
        if line:
            imports.add(line)
    imports -= set(STANDARD_PACKAGES)
    imports -= set((package,))
    for item in sorted(imports):
        print(item)

@task
def find_package(ctx, name):
    ctx.run('pip freeze | grep -i {}'.format(name))

####################################################################################################

@task()
def update_git_sha(ctx):
    # Fixme: wrong workflow, must tag the last commit
    result = ctx.run('git describe --tags --abbrev=0 --always', hide='out')
    tag = result.stdout.strip()
    if tag.startswith('v'):
        version = tag[1:]
    else:
        version = tag
    if not re.match('\d+(\.\d+(\.\d+)?)?', version):
        raise ValueError('Invalid version {}'.format(version))
    result = ctx.run('git rev-parse HEAD', hide='out')
    sha = result.stdout.strip()
    print(sha)
    print(tag)
    print(version)
    filename = Path(ctx.Package, '__init__.py')
    with open(str(filename), 'r') as fh:
        lines = fh.readlines()
    with open(filename, 'w') as fh:
        for line in lines:
            if line.startswith('__version__'):
                line = "__version__ = '{}'\n".format(version)
            if line.startswith('__git_tag__'):
                line = "__git_tag__ = '{}'\n".format(tag)
            if line.startswith('__git_sha__'):
                line = "__git_sha__ = '{}'\n".format(sha)
            fh.write(line)

####################################################################################################

@task()
def clean(ctx):
    for directory in ('build', 'dist', 'PySpice.egg-info'):
        shutil.rmtree(directory, ignore_errors=True)

@task()
def show_python_site(ctx):
    ctx.run('python3 -m site')

@task(update_git_sha)
def build(ctx):
    ctx.run('python3 setup.py build')

@task(build)
def install(ctx):
    ctx.run('python3 setup.py install')

# @task(clean, build)
# def sdist(ctx):
#     ctx.run('python3 setup.py sdist')

# @task(clean, build)
# def bdist(ctx):
#     ctx.run('python3 setup.py bdist')

@task(clean, build)
def wheel(ctx):
    ctx.run('python3 setup.py sdist')
    ctx.run('python3 setup.py bdist_wheel')

@task(wheel)
def upload(ctx):
    # Twine documentation: https://pypi.org/project/twine/
    # ctx.run('twine register dist/*whl')
    # Sign using
    #   BA24CE0F65CB8C67 Fabrice SALVAIRE <gpg AT fabrice-salvaire.fr>
    #   registered on key servers: hkps://hkps.pool.sks-keyservers.net hkp://pgp.mit.edu
    ctx.run('gpg --detach-sign --armor dist/*tar.gz')
    ctx.run('gpg --detach-sign --armor dist/*whl')
    ctx.run('twine upload dist/*')

def _get_pipy_json():
    import requests
    response = requests.get('https://pypi.org/pypi/PySpice/json')
    assert(response.status_code == requests.codes.ok)
    return response.json()

@task()
def get_pypi_json(ctx):
    import json
    print(json.dumps(_get_pipy_json(), sort_keys=True, indent=4))

@task()
def get_wheel(ctx):
    # https://dzone.com/articles/package-signing-in-pip-it-works-in-a-roundabout-so
    data = _get_pipy_json()
    version = data['info']['version']
    filename = data['urls'][0]['filename']
    wheel_url = data['urls'][0]['url']
    print('Get version {}'.format(version))
    ctx.run('curl --output {} {}'.format(filename, wheel_url))
    asc_suffix = '.asc'
    filename_asc = filename + asc_suffix
    ctx.run('curl --output {} {}'.format(filename_asc, wheel_url + asc_suffix))
    # ctx.run('gpg --keyserver pgp.mit.edu --search-keys {}'.format(key_id))
    ctx.run('gpg --verify {} {}'.format(filename_asc, filename))

####################################################################################################

@task()
def get_github_tar_sha(ctx):
    # Fixme: check git sha
    result = ctx.run('git describe --tags --abbrev=0 --always', hide='out')
    tag = result.stdout.strip()
    url = 'https://github.com/FabriceSalvaire/PySpice/archive/{}.tar.gz'.format(tag)
    print('Get', url)
    import hashlib
    import requests
    response = requests.get(url, allow_redirects=True)
    assert(response.status_code == requests.codes.ok)
    sha = hashlib.sha256(response.content)
    print(sha.hexdigest())




####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2017 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

# Note: This module should be outsourced, only code specific to SPICE must remain.

"""This module implements units.

Shortcuts are defined to build unit values easily :

 * for each unit prefix, e.g. :func:`pico`, :func:`nano`, :func:`micro`, :func:`milli`, :func:`kilo`,
   :func:`mega`, :func:`tera`. These shortcuts return unit less values.

 * for each unit and prefix as the concatenation of *u_*, the unit prefix and the
   unit suffix, e.g. :func:`u_pV`, :func:`u_nV`, :func:`u_uV` :func:`u_mV`, :func:`u_V`,
   :func:`u_kV`, :func:`u_MV`, :func:`u_TV`.

Theses unit value constructors accept int, float, object that can be converted to float,
:class:`UnitValue` instance and an iterable on these types.

A shortcut is defined to check an unit value match a particular unit, e.g. :func:`as_V`.  Theses
shortcuts return the value if the unit match else it raises the exception *UnitError*.

A shortcut is defined to access each unit, e.g. :func:`U_V`, :func:`U_A`, :func:`U_s`, :func:`U_Hz`,
:func:`U_Ω`, :func:`U_F`, :func:`U_H.`, as well as for prefixes e.g. :func:`U_mV`.

Some shortcuts have Unicode and ASCII variants:

 * For micro, we have the prefix *μ* and *u*.
 * For Ohm, we have :func:`u_Ω` and :func:`u_Ohm`.

Some examples of usage:

.. code-block:: python3

  foo = kilo(1) # unit less

  resistance_unit = U_Ω

  resistance1 = u_kΩ(1)
  resistance1 = u_kOhm(1) # ASCII variant

  resistance1 = 1@u_kΩ   # using Python 3.5 syntax
  resistance1 = 1 @u_kΩ  # space doesn't matter
  resistance1 = 1 @ u_kΩ #

  resistance2 = as_Ω(resistance1) # check unit

  resistances = u_kΩ(range(1, 11)) # same as [u_kΩ(x) for x in range(1, 11)]
  resistances = range(1, 11)@u_kΩ  # using Python 3.5 syntax

  capacitance = u_uF(200)
  inductance = u_mH(1)
  temperature = u_Degree(25)

  voltage = resistance1 * u_mA(1) # compute unit

  frequency = u_ms(20).frequency
  period = u_Hz(50).period
  pulsation = frequency.pulsation
  pulsation = period.pulsation

.. warning::

   According to the Python `operator precedence
  <https://docs.python.org/3/reference/expressions.html#operator-precedence>`_, division operators
  have a higher priority than the matrix multiplication operator.  In consequence you must had
  parenthesis to perform something like :code:`(10@u_s) / (2@_us)`.

"""

####################################################################################################

import logging
import sys

from . import Unit as _Unit
from . import SiUnits as _SiUnits

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

_version_info = sys.version_info
_has_matmul = _version_info.major * 10 + _version_info.minor >= 35
if not _has_matmul:
    _module_logger.warning("Your Python version doesn't implement @ operator")

####################################################################################################

class UnitValueShorcut:

    ##############################################

    def __init__(self, prefixed_unit):

        self._prefixed_unit = prefixed_unit

    ##############################################

    def _new_value(self, other):

        return self._prefixed_unit.new_value(other)

    ##############################################

    def __call__(self, other):

        """self(other)"""

        return self._new_value(other)

    ##############################################

    def __rmatmul__(self, other):

        """other @ self"""

        return self._new_value(other)

####################################################################################################

def _to_ascii(name):
    ascii_name = name
    for args in (
            ('μ', 'u'),
            ('Ω', 'Ohm'),
            ('°C', 'Degree'),
    ):
        ascii_name = ascii_name.replace(*args)
    return ascii_name

def define_shortcut(name, shortcut) :
    # ° is illegal in Python 3.5
    #  see https://docs.python.org/3/reference/lexical_analysis.html#identifiers
    #      https://www.python.org/dev/peps/pep-3131/
    if '°' not in name:
        globals()[name] = shortcut
    ascii_name = _to_ascii(name)
    if ascii_name != name:
        globals()[ascii_name] = shortcut

####################################################################################################

# Define shortcuts for unit prefixes : ..., micro, milli, kilo, mega, ...

def _build_prefix_shortcut(unit_prefix):
    unit_cls_name = unit_prefix.__class__.__name__
    name = unit_cls_name.lower()
    prefixed_unit = _Unit.PrefixedUnit(power=unit_prefix)
    _Unit.PrefixedUnit.register(prefixed_unit)
    shortcut = lambda value: _Unit.UnitValue(prefixed_unit, value)
    define_shortcut(name, shortcut)

for unit_prefix in _Unit.UnitPrefixMetaclass.prefix_iter():
    if unit_prefix.__class__ != _Unit.ZeroPower:
        _build_prefix_shortcut(unit_prefix) # capture unit_prefix

####################################################################################################

# Fixme: better ???

class FrequencyValue(_Unit.UnitValue, _Unit.FrequencyMixin):
    pass

# Fixme:
class FrequencyValues(_Unit.UnitValues): # , _Unit.FrequencyMixin
    pass

class PeriodValue(_Unit.UnitValue, _Unit.PeriodMixin):
    pass

class PeriodValues(_Unit.UnitValues): # , _Unit.PeriodMixin
    pass

####################################################################################################

# Define unit shortcuts

def _build_unit_type_shortcut(unit):
    name = 'U_' + unit.unit_suffix
    define_shortcut(name, unit)

def _build_as_unit_shortcut(unit):
    name = 'as_' + unit.unit_suffix
    shortcut = unit.validate
    define_shortcut(name, shortcut)

def _exec_body(ns, unit_prefix):
    ns['POWER'] = unit_prefix

def _build_unit_prefix_shortcut(unit, unit_prefix):
    name = 'u_' + str(unit_prefix) + unit.unit_suffix
    if unit.__class__ == _SiUnits.Hertz:
        value_ctor = FrequencyValue
        values_ctor = FrequencyValues
    elif unit.__class__ == _SiUnits.Second:
        value_ctor = PeriodValue
        values_ctor = PeriodValues
    else:
        value_ctor = _Unit.UnitValue
        values_ctor = _Unit.UnitValues
    prefixed_unit = _Unit.PrefixedUnit(unit, unit_prefix, value_ctor, values_ctor)
    _Unit.PrefixedUnit.register(prefixed_unit)
    define_shortcut('U' + name[1:], prefixed_unit)
    shortcut = UnitValueShorcut(prefixed_unit)
    define_shortcut(name, shortcut)

def _build_unit_shortcut(unit):
    _build_as_unit_shortcut(unit)
    _build_unit_type_shortcut(unit)
    for unit_prefix in _Unit.UnitPrefixMetaclass.prefix_iter():
        if unit_prefix.is_defined_in_spice:
            _build_unit_prefix_shortcut(unit, unit_prefix)

for unit in _Unit.UnitMetaclass.unit_iter():
    if unit.unit_suffix and unit.__class__ not in (_SiUnits.Kilogram,):
        # Fixme: kilogram
        _build_unit_shortcut(unit)

####################################################################################################

unit_value = _Unit.UnitValue.simple_value

Frequency = u_Hz
Period = u_s
